################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import warnings
from enum import Enum
from typing import Dict, Union, List, Optional

from pyflink.common import DeserializationSchema, ConfigOptions, Duration, SerializationSchema, \
    ConfigOption
from pyflink.datastream.connectors import Source, Sink, DeliveryGuarantee
from pyflink.java_gateway import get_gateway
from pyflink.util.java_utils import load_java_class


__all__ = [
    'PulsarSource',
    'PulsarSourceBuilder',
    'StartCursor',
    'StopCursor',
    'RangeGenerator',
    'PulsarSink',
    'PulsarSinkBuilder',
    'MessageDelayer',
    'TopicRoutingMode'
]


# ---- PulsarSource ----

class StartCursor(object):
    """
    A factory class for users to specify the start position of a pulsar subscription.
    Since it would be serialized into split.
    The implementation for this interface should be well considered.
    I don't recommend adding extra internal state for this implementation.

    This class would be used only for SubscriptionType.Exclusive and SubscriptionType.Failover.
    """

    def __init__(self, _j_start_cursor):
        self._j_start_cursor = _j_start_cursor

    @staticmethod
    def default_start_cursor() -> 'StartCursor':
        return StartCursor.earliest()

    @staticmethod
    def earliest() -> 'StartCursor':
        JStartCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StartCursor
        return StartCursor(JStartCursor.earliest())

    @staticmethod
    def latest() -> 'StartCursor':
        JStartCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StartCursor
        return StartCursor(JStartCursor.latest())

    @staticmethod
    def from_message_id(message_id: bytes, inclusive: bool = True) -> 'StartCursor':
        """
        Find the available message id and start consuming from it. User could call pulsar Python
        library serialize method to cover messageId bytes.

        Example:
        ::

            >>> from pulsar import MessageId
            >>> message_id_bytes = MessageId().serialize()
            >>> start_cursor = StartCursor.from_message_id(message_id_bytes)
        """
        JStartCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StartCursor
        j_message_id = get_gateway().jvm.org.apache.pulsar.client.api.MessageId \
            .fromByteArray(message_id)
        return StartCursor(JStartCursor.fromMessageId(j_message_id, inclusive))

    @staticmethod
    def from_publish_time(timestamp: int) -> 'StartCursor':
        """
        Seek the start position by using message publish time.
        """
        JStartCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StartCursor
        return StartCursor(JStartCursor.fromPublishTime(timestamp))


class StopCursor(object):
    """
    A factory class for users to specify the stop position of a pulsar subscription. Since it would
    be serialized into split. The implementation for this interface should be well considered. I
    don't recommend adding extra internal state for this implementation.
    """

    def __init__(self, _j_stop_cursor):
        self._j_stop_cursor = _j_stop_cursor

    @staticmethod
    def default_stop_cursor() -> 'StopCursor':
        return StopCursor.never()

    @staticmethod
    def never() -> 'StopCursor':
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        return StopCursor(JStopCursor.never())

    @staticmethod
    def latest() -> 'StopCursor':
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        return StopCursor(JStopCursor.latest())

    @staticmethod
    def at_event_time(timestamp: int) -> 'StopCursor':
        """
        Stop consuming when message eventTime is greater than or equals the specified timestamp.
        """
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        return StopCursor(JStopCursor.atEventTime(timestamp))

    @staticmethod
    def after_event_time(timestamp: int) -> 'StopCursor':
        """
        Stop consuming when message eventTime is greater than the specified timestamp.
        """
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        return StopCursor(JStopCursor.afterEventTime(timestamp))

    @staticmethod
    def at_publish_time(timestamp: int) -> 'StopCursor':
        """
        Stop consuming when message publishTime is greater than or equals the specified timestamp.
        """
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        return StopCursor(JStopCursor.atPublishTime(timestamp))

    @staticmethod
    def after_publish_time(timestamp: int) -> 'StopCursor':
        """
        Stop consuming when message publishTime is greater than the specified timestamp.
        """
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        return StopCursor(JStopCursor.afterPublishTime(timestamp))

    @staticmethod
    def at_message_id(message_id: bytes) -> 'StopCursor':
        """
        Stop consuming when the messageId is equal or greater than the specified messageId.
        Message that is equal to the specified messageId will not be consumed. User could call
        pulsar Python library serialize method to cover messageId bytes.

        Example:
        ::

            >>> from pulsar import MessageId
            >>> message_id_bytes = MessageId().serialize()
            >>> stop_cursor = StopCursor.at_message_id(message_id_bytes)
        """
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        j_message_id = get_gateway().jvm.org.apache.pulsar.client.api.MessageId \
            .fromByteArray(message_id)
        return StopCursor(JStopCursor.atMessageId(j_message_id))

    @staticmethod
    def after_message_id(message_id: bytes) -> 'StopCursor':
        """
        Stop consuming when the messageId is greater than the specified messageId. Message that is
        equal to the specified messageId will be consumed. User could call pulsar Python library
        serialize method to cover messageId bytes.

        Example:
        ::

            >>> from pulsar import MessageId
            >>> message_id_bytes = MessageId().serialize()
            >>> stop_cursor = StopCursor.after_message_id(message_id_bytes)
        """
        JStopCursor = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.cursor.StopCursor
        j_message_id = get_gateway().jvm.org.apache.pulsar.client.api.MessageId \
            .fromByteArray(message_id)
        return StopCursor(JStopCursor.afterMessageId(j_message_id))


class RangeGenerator(object):
    """
    A generator for generating the TopicRange for given topic. It was used for pulsar's
    SubscriptionType#Key_Shared mode. TopicRange would be used in KeySharedPolicy for different
    pulsar source readers.

    If you implement this interface, make sure that each TopicRange would be assigned to a
    specified source reader. Since flink parallelism is provided, make sure the pulsar message key's
    hashcode is evenly distributed among these topic ranges.
    """

    def __init__(self, j_range_generator):
        self._j_range_generator = j_range_generator

    @staticmethod
    def full() -> 'RangeGenerator':
        """
        Default implementation for SubscriptionType#Shared, SubscriptionType#Failover and
        SubscriptionType#Exclusive subscription.
        """
        JFullRangeGenerator = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.topic.range.FullRangeGenerator
        return RangeGenerator(JFullRangeGenerator())

    @staticmethod
    def fixed_key(support_null_key: bool = False,
                  keys: Optional[Union[str, List[str]]] = None,
                  key_bytes: Optional[bytes] = None,
                  ordering_key_bytes: Optional[bytes] = None) -> 'RangeGenerator':
        """
        Pulsar didn't expose the key hash range method. We have to provide an implementation for
        end-user. You can add the keys you want to consume, no need to provide any hash ranges.

        Since the key's hash isn't specified to only one key. The consuming results may contain the
        messages with different keys comparing the keys you have defined in this range generator.
        Remember to use flink's DataStream.filter() method.

        :param support_null_key: Some Message in Pulsar may not have Message#getOrderingKey() or
                                 Message#getKey(), use this method for supporting consuming such
                                 messages.
        :param keys: If you set the message key by using PulsarMessageBuilder#key(String) or
                     TypedMessageBuilder#key(String), use this method for supporting consuming such
                     messages.
        :param key_bytes: If you set the message key by using TypedMessageBuilder#keyBytes(byte[]),
                          use this method for supporting consuming such messages.
        :param ordering_key_bytes: Pulsar's ordering key is prior to the message key. If you set
                                   the ordering key by using
                                   PulsarMessageBuilder#orderingKey(byte[]) or
                                   TypedMessageBuilder#orderingKey(byte[]), use this method for
                                   supporting consuming such messages.
         * messages.
        """
        JFixedKeysRangeGenerator = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.source.enumerator.topic.range.FixedKeysRangeGenerator
        j_range_generator_builder = JFixedKeysRangeGenerator.builder()
        if support_null_key:
            j_range_generator_builder.supportNullKey()
        if keys is not None:
            if isinstance(keys, str):
                j_range_generator_builder.key(keys)
            else:
                for key in keys:
                    j_range_generator_builder.key(key)
        if key_bytes is not None:
            j_range_generator_builder.keyBytes(key_bytes)
        if ordering_key_bytes is not None:
            j_range_generator_builder.orderingKey(ordering_key_bytes)
        return RangeGenerator(j_range_generator_builder.build())


class PulsarSource(Source):
    """
    The Source implementation of Pulsar. Please use a PulsarSourceBuilder to construct a
    PulsarSource. The following example shows how to create a PulsarSource emitting records of
    String type.

    Example:
    ::

        >>> source = PulsarSource() \\
        ...     .builder() \\
        ...     .set_topics([TOPIC1, TOPIC2]) \\
        ...     .set_service_url(get_service_url()) \\
        ...     .set_admin_url(get_admin_url()) \\
        ...     .set_subscription_name("test") \\
        ...     .set_deserialization_schema(SimpleStringSchema()) \\
        ...     .set_bounded_stop_cursor(StopCursor.default_stop_cursor()) \\
        ...     .build()

    See PulsarSourceBuilder for more details.
    """

    def __init__(self, j_pulsar_source):
        super(PulsarSource, self).__init__(source=j_pulsar_source)

    @staticmethod
    def builder() -> 'PulsarSourceBuilder':
        """
        Get a PulsarSourceBuilder to builder a PulsarSource.
        """
        return PulsarSourceBuilder()


class PulsarSourceBuilder(object):
    """
    The builder class for PulsarSource to make it easier for the users to construct a PulsarSource.

    The following example shows the minimum setup to create a PulsarSource that reads the String
    values from a Pulsar topic.

    Example:
    ::

        >>> source = PulsarSource() \\
        ...     .builder() \\
        ...     .set_service_url(PULSAR_BROKER_URL) \\
        ...     .set_admin_url(PULSAR_BROKER_HTTP_URL) \\
        ...     .set_subscription_name("flink-source-1") \\
        ...     .set_topics([TOPIC1, TOPIC2]) \\
        ...     .set_deserialization_schema(SimpleStringSchema()) \\
        ...     .build()

    The service url, admin url, subscription name, topics to consume, and the record deserializer
    are required fields that must be set.

    To specify the starting position of PulsarSource, one can call set_start_cursor(StartCursor).

    By default the PulsarSource runs in an Boundedness.CONTINUOUS_UNBOUNDED mode and never stop
    until the Flink job is canceled or fails. To let the PulsarSource run in
    Boundedness.CONTINUOUS_UNBOUNDED but stops at some given offsets, one can call
    set_unbounded_stop_cursor(StopCursor).

    For example the following PulsarSource stops after it consumes up to a event time when the
    Flink started.

    Example:
    ::

        >>> source = PulsarSource() \\
        ...     .builder() \\
        ...     .set_service_url(PULSAR_BROKER_URL) \\
        ...     .set_admin_url(PULSAR_BROKER_HTTP_URL) \\
        ...     .set_subscription_name("flink-source-1") \\
        ...     .set_topics([TOPIC1, TOPIC2]) \\
        ...     .set_deserialization_schema(SimpleStringSchema()) \\
        ...     .set_bounded_stop_cursor(StopCursor.at_publish_time(int(time.time() * 1000)))
        ...     .build()
    """

    def __init__(self):
        JPulsarSource = \
            get_gateway().jvm.org.apache.flink.connector.pulsar.source.PulsarSource
        self._j_pulsar_source_builder = JPulsarSource.builder()

    def set_admin_url(self, admin_url: str) -> 'PulsarSourceBuilder':
        """
        Sets the admin endpoint for the PulsarAdmin of the PulsarSource.
        """
        self._j_pulsar_source_builder.setAdminUrl(admin_url)
        return self

    def set_service_url(self, service_url: str) -> 'PulsarSourceBuilder':
        """
        Sets the server's link for the PulsarConsumer of the PulsarSource.
        """
        self._j_pulsar_source_builder.setServiceUrl(service_url)
        return self

    def set_subscription_name(self, subscription_name: str) -> 'PulsarSourceBuilder':
        """
        Sets the name for this pulsar subscription.
        """
        self._j_pulsar_source_builder.setSubscriptionName(subscription_name)
        return self

    def set_topics(self, topics: Union[str, List[str]]) -> 'PulsarSourceBuilder':
        """
        Set a pulsar topic list for flink source. Some topic may not exist currently, consuming this
        non-existed topic wouldn't throw any exception. But the best solution is just consuming by
        using a topic regex. You can set topics once either with setTopics or setTopicPattern in
        this builder.
        """
        if not isinstance(topics, list):
            topics = [topics]
        self._j_pulsar_source_builder.setTopics(topics)
        return self

    def set_topic_pattern(self, topic_pattern: str) -> 'PulsarSourceBuilder':
        """
        Set a topic pattern to consume from the java regex str. You can set topics once either with
        set_topics or set_topic_pattern in this builder.
        """
        self._j_pulsar_source_builder.setTopicPattern(topic_pattern)
        return self

    def set_consumer_name(self, consumer_name: str) -> 'PulsarSourceBuilder':
        """
        The consumer name is informative, and it can be used to identify a particular consumer
        instance from the topic stats.

        .. versionadded:: 1.17.2
        """
        self._j_pulsar_source_builder.setConsumerName(consumer_name)
        return self

    def set_range_generator(self, range_generator: RangeGenerator) -> 'PulsarSourceBuilder':
        """
        Set a topic range generator for consuming a sub set of keys.

        :param range_generator: A generator which would generate a set of TopicRange for given
                                topic.

        .. versionadded:: 1.17.2
        """
        self._j_pulsar_source_builder.setRangeGenerator(range_generator._j_range_generator)
        return self

    def set_start_cursor(self, start_cursor: StartCursor) -> 'PulsarSourceBuilder':
        """
        Specify from which offsets the PulsarSource should start consume from by providing an
        StartCursor.
        """
        self._j_pulsar_source_builder.setStartCursor(start_cursor._j_start_cursor)
        return self

    def set_unbounded_stop_cursor(self, stop_cursor: StopCursor) -> 'PulsarSourceBuilder':
        """
        By default the PulsarSource is set to run in Boundedness.CONTINUOUS_UNBOUNDED manner and
        thus never stops until the Flink job fails or is canceled. To let the PulsarSource run as a
        streaming source but still stops at some point, one can set an StopCursor to specify the
        stopping offsets for each partition. When all the partitions have reached their stopping
        offsets, the PulsarSource will then exit.

        This method is different from set_bounded_stop_cursor(StopCursor) that after setting the
        stopping offsets with this method, PulsarSource.getBoundedness() will still return
        Boundedness.CONTINUOUS_UNBOUNDED even though it will stop at the stopping offsets specified
        by the stopping offsets StopCursor.
        """
        self._j_pulsar_source_builder.setUnboundedStopCursor(stop_cursor._j_stop_cursor)
        return self

    def set_bounded_stop_cursor(self, stop_cursor: StopCursor) -> 'PulsarSourceBuilder':
        """
        By default the PulsarSource is set to run in Boundedness.CONTINUOUS_UNBOUNDED manner and
        thus never stops until the Flink job fails or is canceled. To let the PulsarSource run in
        Boundedness.BOUNDED manner and stops at some point, one can set an StopCursor to specify
        the stopping offsets for each partition. When all the partitions have reached their stopping
        offsets, the PulsarSource will then exit.

        This method is different from set_unbounded_stop_cursor(StopCursor) that after setting the
        stopping offsets with this method, PulsarSource.getBoundedness() will return
        Boundedness.BOUNDED instead of Boundedness.CONTINUOUS_UNBOUNDED.
        """
        self._j_pulsar_source_builder.setBoundedStopCursor(stop_cursor._j_stop_cursor)
        return self

    def set_deserialization_schema(self, deserialization_schema: DeserializationSchema) \
            -> 'PulsarSourceBuilder':
        """
        Sets the :class:`~pyflink.common.serialization.DeserializationSchema` for deserializing the
        value of Pulsars message.

        :param deserialization_schema: the :class:`DeserializationSchema` to use for
            deserialization.
        :return: this PulsarSourceBuilder.
        """
        self._j_pulsar_source_builder.setDeserializationSchema(
            deserialization_schema._j_deserialization_schema)
        return self

    def set_authentication(self,
                           auth_plugin_class_name: str,
                           auth_params_string: Union[str, Dict[str, str]]) \
            -> 'PulsarSourceBuilder':
        """
        Configure the authentication provider to use in the Pulsar client instance.

        :param auth_plugin_class_name: Name of the Authentication-Plugin you want to use.
        :param auth_params_string: String which represents parameters for the Authentication-Plugin,
                                   e.g., "key1:val1,key2:val2".

        .. versionadded:: 1.17.2
        """
        if isinstance(auth_params_string, str):
            self._j_pulsar_source_builder.setAuthentication(
                auth_plugin_class_name, auth_params_string)
        else:
            j_auth_params_map = get_gateway().jvm.java.util.HashMap()
            for k, v in auth_params_string.items():
                j_auth_params_map.put(k, v)
            self._j_pulsar_source_builder.setAuthentication(
                auth_plugin_class_name, j_auth_params_map)
        return self

    def set_config(self, key: Union[str, ConfigOption], value) -> 'PulsarSourceBuilder':
        """
        Set arbitrary properties for the PulsarSource and PulsarConsumer. The valid keys can be
        found in PulsarSourceOptions and PulsarOptions.

        Make sure the option could be set only once or with same value.
        """
        if isinstance(key, ConfigOption):
            warnings.warn("set_config(key: ConfigOption, value) is deprecated. "
                          "Use set_config(key: str, value) instead.",
                          DeprecationWarning, stacklevel=2)
            j_config_option = key._j_config_option
        else:
            j_config_option = \
                ConfigOptions.key(key).string_type().no_default_value()._j_config_option
        self._j_pulsar_source_builder.setConfig(j_config_option, value)
        return self

    def set_config_with_dict(self, config: Dict) -> 'PulsarSourceBuilder':
        """
        Set arbitrary properties for the PulsarSource and PulsarConsumer. The valid keys can be
        found in PulsarSourceOptions and PulsarOptions.
        """
        warnings.warn("set_config_with_dict is deprecated. Use set_properties instead.",
                      DeprecationWarning, stacklevel=2)
        self.set_properties(config)
        return self

    def set_properties(self, config: Dict) -> 'PulsarSourceBuilder':
        """
        Set arbitrary properties for the PulsarSource and PulsarConsumer. The valid keys can be
        found in PulsarSourceOptions and PulsarOptions.
        """
        JConfiguration = get_gateway().jvm.org.apache.flink.configuration.Configuration
        self._j_pulsar_source_builder.setConfig(JConfiguration.fromMap(config))
        return self

    def build(self) -> 'PulsarSource':
        """
        Build the PulsarSource.
        """
        return PulsarSource(self._j_pulsar_source_builder.build())


# ---- PulsarSink ----

class TopicRoutingMode(Enum):
    """
    The routing policy for choosing the desired topic by the given message.

    :data: `ROUND_ROBIN`:

    The producer will publish messages across all partitions in a round-robin fashion to achieve
    maximum throughput. Please note that round-robin is not done per individual message but
    rather it's set to the same boundary of batching delay, to ensure batching is effective.

    :data: `MESSAGE_KEY_HASH`:

    If no key is provided, The partitioned producer will randomly pick one single topic partition
    and publish all the messages into that partition. If a key is provided on the message, the
    partitioned producer will hash the key and assign the message to a particular partition.

    :data: `CUSTOM`:

    Use custom topic router implementation that will be called to determine the partition for a
    particular message.
    """

    ROUND_ROBIN = 0
    MESSAGE_KEY_HASH = 1
    CUSTOM = 2

    def _to_j_topic_routing_mode(self):
        JTopicRoutingMode = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.sink.writer.router.TopicRoutingMode
        return getattr(JTopicRoutingMode, self.name)


class MessageDelayer(object):
    """
    A delayer for Pulsar broker passing the sent message to the downstream consumer. This is only
    works in :data:`SubscriptionType.Shared` subscription.

    Read delayed message delivery
    https://pulsar.apache.org/docs/en/next/concepts-messaging/#delayed-message-delivery for better
    understanding this feature.
    """
    def __init__(self, _j_message_delayer):
        self._j_message_delayer = _j_message_delayer

    @staticmethod
    def never() -> 'MessageDelayer':
        """
        All the messages should be consumed immediately.
        """
        JMessageDelayer = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.sink.writer.delayer.MessageDelayer
        return MessageDelayer(JMessageDelayer.never())

    @staticmethod
    def fixed(duration: Duration) -> 'MessageDelayer':
        """
        All the messages should be consumed in a fixed duration.
        """
        JMessageDelayer = get_gateway().jvm \
            .org.apache.flink.connector.pulsar.sink.writer.delayer.MessageDelayer
        return MessageDelayer(JMessageDelayer.fixed(duration._j_duration))


class PulsarSink(Sink):
    """
    The Sink implementation of Pulsar. Please use a PulsarSinkBuilder to construct a
    PulsarSink. The following example shows how to create a PulsarSink receiving records of
    String type.

    Example:
    ::

        >>> sink = PulsarSink.builder() \\
        ...     .set_service_url(PULSAR_BROKER_URL) \\
        ...     .set_admin_url(PULSAR_BROKER_HTTP_URL) \\
        ...     .set_topics(topic) \\
        ...     .set_serialization_schema(SimpleStringSchema()) \\
        ...     .build()

    The sink supports all delivery guarantees described by DeliveryGuarantee.

    DeliveryGuarantee#NONE does not provide any guarantees: messages may be lost in
    case of issues on the Pulsar broker and messages may be duplicated in case of a Flink
    failure.

    DeliveryGuarantee#AT_LEAST_ONCE the sink will wait for all outstanding records in
    the Pulsar buffers to be acknowledged by the Pulsar producer on a checkpoint. No messages
    will be lost in case of any issue with the Pulsar brokers but messages may be duplicated
    when Flink restarts.

    DeliveryGuarantee#EXACTLY_ONCE: In this mode the PulsarSink will write all messages
    in a Pulsar transaction that will be committed to Pulsar on a checkpoint. Thus, no
    duplicates will be seen in case of a Flink restart. However, this delays record writing
    effectively until a checkpoint is written, so adjust the checkpoint duration accordingly.
    Additionally, it is highly recommended to tweak Pulsar transaction timeout (link) >>
    maximum checkpoint duration + maximum restart duration or data loss may happen when Pulsar
    expires an uncommitted transaction.

    See PulsarSinkBuilder for more details.
    """

    def __init__(self, j_pulsar_sink):
        super(PulsarSink, self).__init__(sink=j_pulsar_sink)

    @staticmethod
    def builder() -> 'PulsarSinkBuilder':
        """
        Get a PulsarSinkBuilder to builder a PulsarSink.
        """
        return PulsarSinkBuilder()


class PulsarSinkBuilder(object):
    """
    The builder class for PulsarSink to make it easier for the users to construct a PulsarSink.

    The following example shows the minimum setup to create a PulsarSink that reads the String
    values from a Pulsar topic.

    Example:
    ::

        >>> sink = PulsarSink.builder() \\
        ...     .set_service_url(PULSAR_BROKER_URL) \\
        ...     .set_admin_url(PULSAR_BROKER_HTTP_URL) \\
        ...     .set_topics([TOPIC1, TOPIC2]) \\
        ...     .set_serialization_schema(SimpleStringSchema()) \\
        ...     .build()

    The service url, admin url, and the record serializer are required fields that must be set. If
    you don't set the topics, make sure you have provided a custom TopicRouter. Otherwise,
    you must provide the topics to produce.

    To specify the delivery guarantees of PulsarSink, one can call
    #setDeliveryGuarantee(DeliveryGuarantee). The default value of the delivery guarantee is
    DeliveryGuarantee#NONE, and it wouldn't promise the consistence when write the message into
    Pulsar.

    Example:
    ::

        >>> sink = PulsarSink.builder() \\
        ...     .set_service_url(PULSAR_BROKER_URL) \\
        ...     .set_admin_url(PULSAR_BROKER_HTTP_URL) \\
        ...     .set_topics([TOPIC1, TOPIC2]) \\
        ...     .set_serialization_schema(SimpleStringSchema()) \\
        ...     .set_delivery_guarantee(DeliveryGuarantee.EXACTLY_ONCE)
        ...     .build()
    """

    def __init__(self):
        JPulsarSink = get_gateway().jvm.org.apache.flink.connector.pulsar.sink.PulsarSink
        self._j_pulsar_sink_builder = JPulsarSink.builder()

    def set_admin_url(self, admin_url: str) -> 'PulsarSinkBuilder':
        """
        Sets the admin endpoint for the PulsarAdmin of the PulsarSink.
        """
        self._j_pulsar_sink_builder.setAdminUrl(admin_url)
        return self

    def set_service_url(self, service_url: str) -> 'PulsarSinkBuilder':
        """
        Sets the server's link for the PulsarProducer of the PulsarSink.
        """
        self._j_pulsar_sink_builder.setServiceUrl(service_url)
        return self

    def set_producer_name(self, producer_name: str) -> 'PulsarSinkBuilder':
        """
        The producer name is informative, and it can be used to identify a particular producer
        instance from the topic stats.
        """
        self._j_pulsar_sink_builder.setProducerName(producer_name)
        return self

    def set_topics(self, topics: Union[str, List[str]]) -> 'PulsarSinkBuilder':
        """
        Set a pulsar topic list for flink sink. Some topic may not exist currently, write to this
        non-existed topic wouldn't throw any exception.
        """
        if not isinstance(topics, list):
            topics = [topics]
        self._j_pulsar_sink_builder.setTopics(topics)
        return self

    def set_delivery_guarantee(self, delivery_guarantee: DeliveryGuarantee) -> 'PulsarSinkBuilder':
        """
        Sets the wanted the DeliveryGuarantee. The default delivery guarantee is
        DeliveryGuarantee#NONE.
        """
        self._j_pulsar_sink_builder.setDeliveryGuarantee(
            delivery_guarantee._to_j_delivery_guarantee())
        return self

    def set_topic_routing_mode(self, topic_routing_mode: TopicRoutingMode) -> 'PulsarSinkBuilder':
        """
        Set a routing mode for choosing right topic partition to send messages.
        """
        self._j_pulsar_sink_builder.setTopicRoutingMode(
            topic_routing_mode._to_j_topic_routing_mode())
        return self

    def set_topic_router(self, topic_router_class_name: str) -> 'PulsarSinkBuilder':
        """
        Use a custom topic router instead predefine topic routing.
        """
        j_topic_router = load_java_class(topic_router_class_name).newInstance()
        self._j_pulsar_sink_builder.setTopicRouter(j_topic_router)
        return self

    def set_serialization_schema(self, serialization_schema: SerializationSchema) \
            -> 'PulsarSinkBuilder':
        """
        Sets the SerializationSchema of the PulsarSinkBuilder.
        """
        self._j_pulsar_sink_builder.setSerializationSchema(
            serialization_schema._j_serialization_schema)
        return self

    def set_authentication(self,
                           auth_plugin_class_name: str,
                           auth_params_string: Union[str, Dict[str, str]]) \
            -> 'PulsarSinkBuilder':
        """
        Configure the authentication provider to use in the Pulsar client instance.

        :param auth_plugin_class_name: Name of the Authentication-Plugin you want to use.
        :param auth_params_string: String which represents parameters for the Authentication-Plugin,
                                   e.g., "key1:val1,key2:val2".

        .. versionadded:: 1.17.2
        """
        if isinstance(auth_params_string, str):
            self._j_pulsar_sink_builder.setAuthentication(
                auth_plugin_class_name, auth_params_string)
        else:
            j_auth_params_map = get_gateway().jvm.java.util.HashMap()
            for k, v in auth_params_string.items():
                j_auth_params_map.put(k, v)
            self._j_pulsar_sink_builder.setAuthentication(
                auth_plugin_class_name, j_auth_params_map)
        return self

    def delay_sending_message(self, message_delayer: MessageDelayer) -> 'PulsarSinkBuilder':
        """
        Set a message delayer for enable Pulsar message delay delivery.
        """
        self._j_pulsar_sink_builder.delaySendingMessage(message_delayer._j_message_delayer)
        return self

    def set_config(self, key: str, value) -> 'PulsarSinkBuilder':
        """
        Set an arbitrary property for the PulsarSink and Pulsar Producer. The valid keys can be
        found in PulsarSinkOptions and PulsarOptions.

        Make sure the option could be set only once or with same value.
        """
        j_config_option = ConfigOptions.key(key).string_type().no_default_value()._j_config_option
        self._j_pulsar_sink_builder.setConfig(j_config_option, value)
        return self

    def set_properties(self, config: Dict) -> 'PulsarSinkBuilder':
        """
        Set an arbitrary property for the PulsarSink and Pulsar Producer. The valid keys can be
        found in PulsarSinkOptions and PulsarOptions.
        """
        JConfiguration = get_gateway().jvm.org.apache.flink.configuration.Configuration
        self._j_pulsar_sink_builder.setConfig(JConfiguration.fromMap(config))
        return self

    def build(self) -> 'PulsarSink':
        """
        Build the PulsarSink.
        """
        return PulsarSink(self._j_pulsar_sink_builder.build())



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from libc.stdint cimport int64_t
from pyflink.fn_execution.table.aggregate_fast cimport RowKeySelector
from pyflink.fn_execution.coder_impl_fast cimport InternalRow

cdef class NamespaceAggsHandleFunctionBase:
    cdef void open(self, object state_data_view_store)
    cdef void accumulate(self, InternalRow input_data)
    cdef void retract(self, InternalRow input_data)
    cpdef void merge(self, object namespace, list accumulators)
    cpdef void set_accumulators(self, object namespace, list accumulators)
    cdef list get_accumulators(self)
    cpdef list create_accumulators(self)
    cpdef void cleanup(self, object namespace)
    cdef void close(self)

cdef class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase):
    cpdef list get_value(self, object namespace)

cdef class SimpleNamespaceAggsHandleFunction(NamespaceAggsHandleFunction):
    cdef list _udfs
    cdef size_t _udf_num
    cdef list _input_extractors
    cdef object _named_property_extractor
    cdef list _accumulators
    cdef list _udf_data_view_specs
    cdef list _udf_data_views
    cdef int*_filter_args
    cdef int*_distinct_indexes
    cdef dict _distinct_view_descriptors
    cdef dict _distinct_data_views
    cdef object named_property_extractor
    cdef size_t*_get_value_indexes
    cdef size_t _get_value_indexes_length

cdef class GroupWindowAggFunctionBase:
    cdef int64_t _allowed_lateness
    cdef int _rowtime_index
    cdef str _shift_timezone
    cdef RowKeySelector _key_selector
    cdef object _state_value_coder
    cdef object _state_backend
    cdef object _window_assigner
    cdef NamespaceAggsHandleFunctionBase _window_aggregator
    cdef object _trigger
    cdef object _window_function
    cdef object _internal_timer_service
    cdef object _window_context
    cdef object _trigger_context
    cdef object _window_state

    cpdef void open(self, object function_context)
    cpdef void close(self)
    cpdef list process_element(self, InternalRow input_data)
    cpdef void process_watermark(self, int64_t watermark)
    cpdef list on_event_time(self, object timer)
    cpdef list on_processing_time(self, object timer)
    cpdef list get_timers(self)
    cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills)
    cpdef int64_t cleanup_time(self, object window)
    cdef void _register_cleanup_timer(self, object window)
    cdef InternalRow _emit_window_result(self, list key, object window)

cdef class GroupWindowAggFunction(GroupWindowAggFunctionBase):
    pass



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import typing
from typing import TypeVar, Iterable, Collection, Optional

from pyflink.common.constants import MAX_LONG_VALUE
from pyflink.common.typeinfo import PickledBytesTypeInfo
from pyflink.datastream import WindowAssigner, Trigger, MergingWindowAssigner, TriggerResult
from pyflink.datastream.functions import KeyedStateStore, RuntimeContext, InternalWindowFunction
from pyflink.datastream.output_tag import OutputTag
from pyflink.datastream.state import StateDescriptor, ListStateDescriptor, \
    ReducingStateDescriptor, AggregatingStateDescriptor, ValueStateDescriptor, MapStateDescriptor, \
    State, AggregatingState, ReducingState, MapState, ListState, ValueState, AppendingState
from pyflink.fn_execution.datastream.timerservice import InternalTimerService
from pyflink.fn_execution.datastream.window.merging_window_set import MergingWindowSet
from pyflink.fn_execution.internal_state import InternalMergingState, InternalKvState, \
    InternalAppendingState
from pyflink.metrics import MetricGroup

T = TypeVar("T")
IN = TypeVar("IN")
OUT = TypeVar("OUT")
KEY = TypeVar("KEY")
W = TypeVar("W")


def get_or_create_keyed_state(runtime_context, state_descriptor):
    if isinstance(state_descriptor, ListStateDescriptor):
        state = runtime_context.get_list_state(state_descriptor)
    elif isinstance(state_descriptor, ReducingStateDescriptor):
        state = runtime_context.get_reducing_state(state_descriptor)
    elif isinstance(state_descriptor, AggregatingStateDescriptor):
        state = runtime_context.get_aggregating_state(state_descriptor)
    elif isinstance(state_descriptor, ValueStateDescriptor):
        state = runtime_context.get_state(state_descriptor)
    elif isinstance(state_descriptor, MapStateDescriptor):
        state = runtime_context.get_map_state(state_descriptor)
    else:
        raise Exception("Unsupported state descriptor: %s" % type(state_descriptor))
    return state


class MergingWindowStateStore(KeyedStateStore):

    def __init__(self):
        self.window = None

    def get_state(self, state_descriptor: ValueStateDescriptor) -> ValueState:
        raise Exception("Per-window state is not allowed when using merging windows.")

    def get_list_state(self, state_descriptor: ListStateDescriptor) -> ListState:
        raise Exception("Per-window state is not allowed when using merging windows.")

    def get_map_state(self, state_descriptor: MapStateDescriptor) -> MapState:
        raise Exception("Per-window state is not allowed when using merging windows.")

    def get_reducing_state(self, state_descriptor: ReducingStateDescriptor) -> ReducingState:
        raise Exception("Per-window state is not allowed when using merging windows.")

    def get_aggregating_state(
            self, state_descriptor: AggregatingStateDescriptor) -> AggregatingState:
        raise Exception("Per-window state is not allowed when using merging windows.")


class PerWindowStateStore(KeyedStateStore):

    def __init__(self, runtime_context):
        self._runtime_context = runtime_context
        self.window = None

    def get_state(self, state_descriptor: ValueStateDescriptor) -> ValueState:
        return self._set_namespace(self._runtime_context.get_state(state_descriptor))

    def get_list_state(self, state_descriptor: ListStateDescriptor) -> ListState:
        return self._set_namespace(self._runtime_context.get_list_state(state_descriptor))

    def get_map_state(self, state_descriptor: MapStateDescriptor) -> MapState:
        return self._set_namespace(self._runtime_context.get_map_state(state_descriptor))

    def get_reducing_state(self, state_descriptor: ReducingStateDescriptor) -> ReducingState:
        return self._set_namespace(self._runtime_context.get_reducing_state(state_descriptor))

    def get_aggregating_state(
            self, state_descriptor: AggregatingStateDescriptor) -> AggregatingState:
        return self._set_namespace(self._runtime_context.get_aggregating_state(state_descriptor))

    def _set_namespace(self, state):
        state.set_current_namespace(self.window)
        return state


class Context(Trigger.OnMergeContext):

    def __init__(
            self,
            runtime_context: RuntimeContext,
            internal_timer_service: InternalTimerService,
            trigger: Trigger):
        self._runtime_context = runtime_context
        self._internal_timer_service = internal_timer_service
        self._trigger = trigger
        self.user_key = None
        self.window = None
        self.merged_windows = None

    def get_current_processing_time(self) -> int:
        return self._internal_timer_service.current_processing_time()

    def get_metric_group(self) -> MetricGroup:
        return self._runtime_context.get_metrics_group()

    def get_current_watermark(self) -> int:
        return self._internal_timer_service.current_watermark()

    def register_processing_time_timer(self, time: int) -> None:
        self._internal_timer_service.register_processing_time_timer(self.window, time)

    def register_event_time_timer(self, time: int) -> None:
        self._internal_timer_service.register_event_time_timer(self.window, time)

    def delete_processing_time_timer(self, time: int) -> None:
        self._internal_timer_service.delete_processing_time_timer(self.window, time)

    def delete_event_time_timer(self, time: int) -> None:
        self._internal_timer_service.delete_event_time_timer(self.window, time)

    def merge_partitioned_state(self, state_descriptor: StateDescriptor) -> None:
        if self.merged_windows is not None and len(self.merged_windows) > 0:
            raw_state = get_or_create_keyed_state(self._runtime_context, state_descriptor)
            if isinstance(raw_state, InternalMergingState):
                raw_state.merge_namespaces(self.window, self.merged_windows)
            else:
                raise Exception(
                    "The given state descriptor does not refer to a mergeable state (MergingState)")

    def get_partitioned_state(self, state_descriptor: StateDescriptor) -> State:
        state = get_or_create_keyed_state(
            self._runtime_context, state_descriptor)  # type: InternalKvState
        state.set_current_namespace(self.window)
        return state

    def on_element(self, value, timestamp) -> TriggerResult:
        return self._trigger.on_element(value, timestamp, self.window, self)

    def on_processing_time(self, time) -> TriggerResult:
        return self._trigger.on_processing_time(time, self.window, self)

    def on_event_time(self, time) -> TriggerResult:
        return self._trigger.on_event_time(time, self.window, self)

    def on_merge(self, merged_windows) -> None:
        self.merged_windows = merged_windows
        self._trigger.on_merge(self.window, self)

    def clear(self) -> None:
        self._trigger.clear(self.window, self)


class WindowContext(InternalWindowFunction.InternalWindowContext):

    def __init__(self,
                 window_assigner: WindowAssigner,
                 runtime_context: RuntimeContext,
                 window_function: InternalWindowFunction,
                 internal_timer_service: InternalTimerService):
        self.window = None
        if isinstance(window_assigner, MergingWindowAssigner):
            self._window_state = MergingWindowStateStore()
        else:
            self._window_state = PerWindowStateStore(runtime_context)
        self._runtime_context = runtime_context
        self._user_function = window_function
        self._internal_timer_service = internal_timer_service

    def current_processing_time(self) -> int:
        return self._internal_timer_service.current_processing_time()

    def current_watermark(self) -> int:
        return self._internal_timer_service.current_watermark()

    def window_state(self) -> KeyedStateStore:
        self._window_state.window = self.window
        return self._window_state

    def global_state(self) -> KeyedStateStore:
        return self._runtime_context

    def clear(self) -> None:
        self._user_function.clear(self.window, self)


class WindowAssignerContext(WindowAssigner.WindowAssignerContext):

    def __init__(self,
                 internal_timer_service: InternalTimerService,
                 runtime_context: RuntimeContext):
        self._internal_timer_service = internal_timer_service
        self._runtime_context = runtime_context

    def get_current_processing_time(self) -> int:
        return self._internal_timer_service.current_processing_time()

    def get_runtime_context(self) -> RuntimeContext:
        return self._runtime_context


class WindowMergeFunction(MergingWindowSet.MergeFunction[W]):

    def __init__(self,
                 window_operator: 'WindowOperator'):
        self._window_assigner = window_operator.window_assigner
        self._internal_timer_service = window_operator.internal_timer_service
        self._allowed_lateness = window_operator.allowed_lateness
        self._trigger_context = window_operator.trigger_context
        self._window_merging_state = window_operator.window_merging_state
        self._user_key_selector = window_operator.user_key_selector
        self.delete_cleanup_timer = window_operator.delete_cleanup_timer

        self.key = None

    def merge(self,
              merge_result: W,
              merged_windows: Collection[W],
              state_window_result: W,
              merged_state_windows: Collection[W]):
        if self._window_assigner.is_event_time() and \
                merge_result.max_timestamp() + self._allowed_lateness <= \
                self._internal_timer_service.current_watermark():
            raise Exception("The end timestamp of an event-time window cannot become earlier than "
                            "the current watermark by merging. Current watermark: %d window: %s" %
                            (self._internal_timer_service.current_watermark(), merge_result))
        elif not self._window_assigner.is_event_time():
            current_processing_time = self._internal_timer_service.current_processing_time()
            if merge_result.max_timestamp() <= current_processing_time:
                raise Exception("The end timestamp of a processing-time window cannot become "
                                "earlier than the current processing time by merging. Current "
                                "processing time: %d window: %s" %
                                (current_processing_time, merge_result))

        self._trigger_context.user_key = self._user_key_selector(self.key)
        self._trigger_context.window = merge_result

        self._trigger_context.on_merge(merged_windows)

        for m in merged_windows:
            self._trigger_context.window = m
            self._trigger_context.clear()
            self.delete_cleanup_timer(m)

        self._window_merging_state.merge_namespaces(state_window_result, merged_state_windows)


class WindowOperator(object):
    LATE_ELEMENTS_DROPPED_METRIC_NAME = "numLateRecordsDropped"

    def __init__(self,
                 window_assigner: WindowAssigner,
                 keyed_state_backend,
                 user_key_selector,
                 window_state_descriptor: StateDescriptor,
                 window_function: InternalWindowFunction,
                 trigger: Trigger,
                 allowed_lateness: int,
                 late_data_output_tag: Optional[OutputTag]):
        self.window_assigner = window_assigner
        self.keyed_state_backend = keyed_state_backend
        self.user_key_selector = user_key_selector
        self.window_state_descriptor = window_state_descriptor
        self.window_function = window_function
        self.trigger = trigger
        self.allowed_lateness = allowed_lateness
        self.late_data_output_tag = late_data_output_tag

        self.num_late_records_dropped = None
        self.internal_timer_service = None  # type: InternalTimerService
        self.trigger_context = None  # type: Context
        self.process_context = None  # type: WindowContext
        self.window_assigner_context = None  # type: WindowAssignerContext
        self.window_state = None  # type: InternalAppendingState
        self.window_merging_state = None  # type: InternalMergingState
        self.merging_sets_state = None

        self.merge_function = None  # type: WindowMergeFunction

    def open(self, runtime_context: RuntimeContext, internal_timer_service: InternalTimerService):
        self.window_function.open(runtime_context)

        self.num_late_records_dropped = runtime_context.get_metrics_group().counter(
            self.LATE_ELEMENTS_DROPPED_METRIC_NAME)
        self.internal_timer_service = internal_timer_service
        self.trigger_context = Context(runtime_context, internal_timer_service, self.trigger)
        self.process_context = WindowContext(
            self.window_assigner,
            runtime_context,
            self.window_function,
            self.internal_timer_service)
        self.window_assigner_context = WindowAssignerContext(
            self.internal_timer_service,
            runtime_context)

        # create (or restore) the state that hold the actual window contents
        # NOTE - the state may be null in the case of the overriding evicting window operator
        if self.window_state_descriptor is not None:
            self.window_state = get_or_create_keyed_state(
                runtime_context, self.window_state_descriptor)

        if isinstance(self.window_assigner, MergingWindowAssigner):
            if isinstance(self.window_state, InternalMergingState):
                self.window_merging_state = self.window_state

            if hasattr(self.keyed_state_backend, 'namespace_coder'):
                window_coder = self.keyed_state_backend.namespace_coder
                self.merging_sets_state = self.keyed_state_backend.get_map_state(
                    "merging-window-set", window_coder, window_coder)
            else:
                state_descriptor = MapStateDescriptor(
                    "merging-window-set",
                    PickledBytesTypeInfo(),
                    PickledBytesTypeInfo())
                self.merging_sets_state = self.keyed_state_backend.get_map_state(state_descriptor)

        self.merge_function = WindowMergeFunction(self)

    def close(self):
        self.window_function.close()
        self.trigger_context = None
        self.process_context = None
        self.window_assigner_context = None

    def process_element(self, value, timestamp: int):
        element_windows = self.window_assigner.assign_windows(
            value, timestamp, self.window_assigner_context)

        is_skipped_element = True

        key = self.keyed_state_backend.get_current_key()
        self.merge_function.key = key

        if isinstance(self.window_assigner, MergingWindowAssigner):
            merging_windows = self.get_merging_window_set()

            for window in element_windows:
                actual_window = merging_windows.add_window(window, self.merge_function)
                if self.is_window_late(actual_window):
                    merging_windows.retire_window(actual_window)
                    continue
                is_skipped_element = False

                state_window = merging_windows.get_state_window(actual_window)
                if state_window is None:
                    raise Exception("Window %s is not in in-flight window set." % state_window)

                self.window_state.set_current_namespace(state_window)
                self.window_state.add(value)

                self.trigger_context.user_key = self.user_key_selector(key)
                self.trigger_context.window = actual_window

                trigger_result = self.trigger_context.on_element(value, timestamp)

                if trigger_result.is_fire():
                    contents = self.window_state.get()
                    # for list state the iterable will never be none
                    if isinstance(self.window_state, ListState):
                        contents = [i for i in contents]
                        if len(contents) == 0:
                            contents = None
                    if contents is None:
                        continue
                    yield from self.emit_window_contents(actual_window, contents)

                if trigger_result.is_purge():
                    self.window_state.clear()
                self.register_cleanup_timer(actual_window)

            merging_windows.persist()
        else:
            for window in element_windows:
                if self.is_window_late(window):
                    continue
                is_skipped_element = False

                self.window_state.set_current_namespace(window)
                self.window_state.add(value)

                self.trigger_context.user_key = self.user_key_selector(key)
                self.trigger_context.window = window

                trigger_result = self.trigger_context.on_element(value, timestamp)

                if trigger_result.is_fire():
                    contents = self.window_state.get()
                    # for list state the iterable will never be none
                    if isinstance(self.window_state, ListState):
                        contents = [i for i in contents]
                        if len(contents) == 0:
                            contents = None
                    if contents is None:
                        continue
                    yield from self.emit_window_contents(window, contents)

                if trigger_result.is_purge():
                    self.window_state.clear()
                self.register_cleanup_timer(window)

        if is_skipped_element and self.is_element_late(value, timestamp):
            if self.late_data_output_tag is not None:
                yield self.late_data_output_tag, value
            else:
                self.num_late_records_dropped.inc()

    def on_event_time(self, timestamp, key, namespace) -> None:
        self.trigger_context.user_key = self.user_key_selector(key)
        self.trigger_context.window = namespace

        if isinstance(self.window_assigner, MergingWindowAssigner):
            merging_windows = self.get_merging_window_set()
            state_window = merging_windows.get_state_window(self.trigger_context.window)
            if state_window is None:
                # Timer firing for non-existent window, this can only happen if a
                # trigger did not clean up timers. We have already cleared the merging
                # window and therefore the Trigger state, however, so nothing to do.
                return
            else:
                self.window_state.set_current_namespace(state_window)
        else:
            self.window_state.set_current_namespace(self.trigger_context.window)
            merging_windows = None

        trigger_result = self.trigger_context.on_event_time(timestamp)

        if trigger_result.is_fire():
            contents = self.window_state.get()
            # for list state the iterable will never be none
            if isinstance(self.window_state, ListState):
                contents = [i for i in contents]
                if len(contents) == 0:
                    contents = None
            if contents is not None:
                yield from self.emit_window_contents(self.trigger_context.window, contents)

        if trigger_result.is_purge():
            self.window_state.clear()

        if self.window_assigner.is_event_time() and self.is_cleanup_time(
                self.trigger_context.window, timestamp):
            self.clear_all_state(self.trigger_context.window, self.window_state, merging_windows)

        if merging_windows is not None:
            merging_windows.persist()

    def on_processing_time(self, timestamp, key, namespace):
        self.trigger_context.user_key = self.user_key_selector(key)
        self.trigger_context.window = namespace

        if isinstance(self.window_assigner, MergingWindowAssigner):
            merging_windows = self.get_merging_window_set()
            state_window = merging_windows.get_state_window(self.trigger_context.window)
            if state_window is None:
                # Timer firing for non-existent window, this can only happen if a
                # trigger did not clean up timers. We have already cleared the merging
                # window and therefore the Trigger state, however, so nothing to do.
                return
            else:
                self.window_state.set_current_namespace(state_window)
        else:
            self.window_state.set_current_namespace(self.trigger_context.window)
            merging_windows = None

        trigger_result = self.trigger_context.on_processing_time(timestamp)

        if trigger_result.is_fire():
            contents = self.window_state.get()
            # for list state the iterable will never be none
            if isinstance(self.window_state, ListState):
                contents = [i for i in contents]
                if len(contents) == 0:
                    contents = None
            if contents is not None:
                yield from self.emit_window_contents(self.trigger_context.window, contents)

        if trigger_result.is_purge():
            self.window_state.clear()

        if not self.window_assigner.is_event_time() and self.is_cleanup_time(
                self.trigger_context.window, timestamp):
            self.clear_all_state(self.trigger_context.window, self.window_state, merging_windows)

        if merging_windows is not None:
            merging_windows.persist()

    def get_merging_window_set(self) -> MergingWindowSet:
        return MergingWindowSet(
            typing.cast(MergingWindowAssigner[T, W], self.window_assigner),
            self.merging_sets_state)

    def cleanup_time(self, window) -> int:
        if self.window_assigner.is_event_time():
            time = window.max_timestamp() + self.allowed_lateness
            if time >= window.max_timestamp():
                return time
            else:
                return MAX_LONG_VALUE
        else:
            return window.max_timestamp()

    def is_cleanup_time(self, window, time) -> bool:
        return time == self.cleanup_time(window)

    def register_cleanup_timer(self, window) -> None:
        cleanup_time = self.cleanup_time(window)
        if cleanup_time == MAX_LONG_VALUE:
            return
        if self.window_assigner.is_event_time():
            self.trigger_context.register_event_time_timer(cleanup_time)
        else:
            self.trigger_context.register_processing_time_timer(cleanup_time)

    def delete_cleanup_timer(self, window) -> None:
        cleanup_time = self.cleanup_time(window)
        if cleanup_time == MAX_LONG_VALUE:
            return
        if self.window_assigner.is_event_time():
            self.trigger_context.delete_event_time_timer(cleanup_time)
        else:
            self.trigger_context.delete_processing_time_timer(cleanup_time)

    def is_window_late(self, window) -> bool:
        return self.window_assigner.is_event_time() and \
            self.cleanup_time(window) <= self.internal_timer_service.current_watermark()

    def is_element_late(self, value, timestamp) -> bool:
        return self.window_assigner.is_event_time() and timestamp + self.allowed_lateness <= \
            self.internal_timer_service.current_watermark()

    def clear_all_state(
            self, window, window_state: AppendingState, merging_windows: MergingWindowSet):
        window_state.clear()
        self.trigger_context.clear()
        self.process_context.window = window
        self.process_context.clear()
        if merging_windows is not None:
            merging_windows.retire_window(window)
            merging_windows.persist()

    def emit_window_contents(self, window, contents) -> Iterable:
        self.process_context.window = window
        return self.window_function.process(
            self.trigger_context.user_key, window, self.process_context, contents)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import abc

from typing import Any, Optional

from pyflink.common.time import Duration
from pyflink.java_gateway import get_gateway


class WatermarkStrategy(object):
    """
    The WatermarkStrategy defines how to generate Watermarks in the stream sources. The
    WatermarkStrategy is a builder/factory for the WatermarkGenerator that generates the watermarks
    and the TimestampAssigner which assigns the internal timestamp of a record.

    The convenience methods, for example forBoundedOutOfOrderness(Duration), create a
    WatermarkStrategy for common built in strategies.
    """
    def __init__(self, j_watermark_strategy):
        self._j_watermark_strategy = j_watermark_strategy
        self._timestamp_assigner = None

    def with_timestamp_assigner(self, timestamp_assigner: 'TimestampAssigner') -> \
            'WatermarkStrategy':
        """
        Creates a new WatermarkStrategy that wraps this strategy but instead uses the given a
        TimestampAssigner by implementing TimestampAssigner interface.

        Example:
        ::

            >>> watermark_strategy = WatermarkStrategy.for_monotonous_timestamps() \\
            >>>     .with_timestamp_assigner(MyTimestampAssigner())

        :param timestamp_assigner: The given TimestampAssigner.
        :return: A WaterMarkStrategy that wraps a TimestampAssigner.
        """
        self._timestamp_assigner = timestamp_assigner
        return self

    def with_idleness(self, idle_timeout: Duration) -> 'WatermarkStrategy':
        """
        Creates a new enriched WatermarkStrategy that also does idleness detection in the created
        WatermarkGenerator.

        Example:
        ::

            >>> WatermarkStrategy \\
            ...     .for_bounded_out_of_orderness(Duration.of_seconds(20)) \\
            ...     .with_idleness(Duration.of_minutes(1))

        :param idle_timeout: The idle timeout.
        :return: A new WatermarkStrategy with idle detection configured.
        """
        return WatermarkStrategy(self._j_watermark_strategy.withIdleness(idle_timeout._j_duration))

    def with_watermark_alignment(self, watermark_group: str, max_allowed_watermark_drift: Duration,
                                 update_interval: Optional[Duration] = None) -> 'WatermarkStrategy':
        """
        Creates a new :class:`WatermarkStrategy` that configures the maximum watermark drift from
        other sources/tasks/partitions in the same watermark group. The group may contain completely
        independent sources (e.g. File and Kafka).

        Once configured Flink will "pause" consuming from a source/task/partition that is ahead
        of the emitted watermark in the group by more than the maxAllowedWatermarkDrift.

        Example:
        ::

            >>> WatermarkStrategy \\
            ...     .for_bounded_out_of_orderness(Duration.of_seconds(20)) \\
            ...     .with_watermark_alignment("alignment-group-1", Duration.of_seconds(20),
            ...         Duration.of_seconds(1))

        :param watermark_group: A group of sources to align watermarks
        :param max_allowed_watermark_drift: Maximal drift, before we pause consuming from the
            source/task/partition
        :param update_interval: How often tasks should notify coordinator about the current
            watermark and how often the coordinator should announce the maximal aligned watermark.
            If is None, default update interval (1000ms) is used.
        :return: A new WatermarkStrategy with watermark alignment configured.

        .. versionadded:: 1.16.0
        """
        if update_interval is None:
            return WatermarkStrategy(self._j_watermark_strategy.withWatermarkAlignment(
                watermark_group, max_allowed_watermark_drift._j_duration
            ))
        else:
            return WatermarkStrategy(self._j_watermark_strategy.withWatermarkAlignment(
                watermark_group,
                max_allowed_watermark_drift._j_duration,
                update_interval._j_duration,
            ))

    @staticmethod
    def for_monotonous_timestamps() -> 'WatermarkStrategy':
        """
        Creates a watermark strategy for situations with monotonously ascending timestamps.

        The watermarks are generated periodically and tightly follow the latest timestamp in the
        data. The delay introduced by this strategy is mainly the periodic interval in which the
        watermarks are generated.
        """
        JWaterMarkStrategy = get_gateway().jvm\
            .org.apache.flink.api.common.eventtime.WatermarkStrategy
        return WatermarkStrategy(JWaterMarkStrategy.forMonotonousTimestamps())

    @staticmethod
    def for_bounded_out_of_orderness(max_out_of_orderness: Duration) -> 'WatermarkStrategy':
        """
        Creates a watermark strategy for situations where records are out of order, but you can
        place an upper bound on how far the events are out of order. An out-of-order bound B means
        that once the an event with timestamp T was encountered, no events older than (T - B) will
        follow any more.
        """
        JWaterMarkStrategy = get_gateway().jvm \
            .org.apache.flink.api.common.eventtime.WatermarkStrategy
        return WatermarkStrategy(
            JWaterMarkStrategy.forBoundedOutOfOrderness(max_out_of_orderness._j_duration))

    @staticmethod
    def no_watermarks() -> 'WatermarkStrategy':
        """
        Creates a watermark strategy that generates no watermarks at all. This may be useful in
        scenarios that do pure processing-time based stream processing.

        .. versionadded:: 1.16.0
        """
        JWaterMarkStrategy = get_gateway().jvm \
            .org.apache.flink.api.common.eventtime.WatermarkStrategy
        return WatermarkStrategy(JWaterMarkStrategy.noWatermarks())


class TimestampAssigner(abc.ABC):
    """
    A TimestampAssigner assigns event time timestamps to elements. These timestamps are used by all
    functions that operate on event time, for example event time windows.

    Timestamps can be an arbitrary int value, but all built-in implementations represent it as the
    milliseconds since the Epoch (midnight, January 1, 1970 UTC), the same way as time.time() does
    it.
    """
    @abc.abstractmethod
    def extract_timestamp(self, value: Any, record_timestamp: int) -> int:
        """
        Assigns a timestamp to an element, in milliseconds since the Epoch. This is independent of
        any particular time zone or calendar.

        The method is passed the previously assigned timestamp of the element.
        That previous timestamp may have been assigned from a previous assigner. If the element did
        not carry a timestamp before, this value is the minimum value of int type.

        :param value: The element that the timestamp will be assigned to.
        :param record_timestamp: The current internal timestamp of the element, or a negative value,
                                 if no timestamp has been assigned yet.
        :return: The new timestamp.
        """
        pass


class AssignerWithPeriodicWatermarksWrapper(object):
    """
    The AssignerWithPeriodicWatermarks assigns event time timestamps to elements, and generates
    low watermarks that signal event time progress within the stream. These timestamps and
    watermarks are used by functions and operators that operate on event time, for example event
    time windows.
    """

    def __init__(self, j_assigner_with_periodic_watermarks):
        self._j_assigner_with_periodic_watermarks = j_assigner_with_periodic_watermarks



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import json
from enum import Enum
from typing import Callable, List, Tuple

from pyflink.fn_execution.metrics.process.counter_impl import CounterImpl
from pyflink.fn_execution.metrics.process.distribution_impl import DistributionImpl
from pyflink.fn_execution.metrics.process.meter_impl import MeterImpl
from pyflink.metrics import MetricGroup, Counter, Distribution, Meter


class MetricGroupType(Enum):
    """
    Indicate the type of MetricGroup.
    """
    generic = 0
    key = 1
    value = 2


class GenericMetricGroup(MetricGroup):

    def __init__(
            self,
            parent,
            name,
            metric_group_type=MetricGroupType.generic):
        self._parent = parent
        self._sub_groups = []
        self._name = name
        self._metric_group_type = metric_group_type
        self._flink_gauge = {}
        self._beam_gauge = {}

    def add_group(self, name: str, extra: str = None) -> 'MetricGroup':
        if extra is None:
            return self._add_group(name, MetricGroupType.generic)
        else:
            return self._add_group(name, MetricGroupType.key) \
                ._add_group(extra, MetricGroupType.value)

    def counter(self, name: str) -> 'Counter':
        from apache_beam.metrics.metric import Metrics
        return CounterImpl(Metrics.counter(self._get_namespace(), name))

    def gauge(self, name: str, obj: Callable[[], int]) -> None:
        from apache_beam.metrics.metric import Metrics
        self._flink_gauge[name] = obj
        self._beam_gauge[name] = Metrics.gauge(self._get_namespace(), name)

    def meter(self, name: str, time_span_in_seconds: int = 60) -> 'Meter':
        from apache_beam.metrics.metric import Metrics
        # There is no meter type in Beam, use counter to implement meter
        return MeterImpl(Metrics.counter(self._get_namespace(time_span_in_seconds), name))

    def distribution(self, name: str) -> 'Distribution':
        from apache_beam.metrics.metric import Metrics
        return DistributionImpl(Metrics.distribution(self._get_namespace(), name))

    def _add_group(self, name: str, metric_group_type: MetricGroupType) -> 'GenericMetricGroup':
        for group in self._sub_groups:
            if name == group._name and metric_group_type == group._metric_group_type:
                # we don't create same metric group repeatedly
                return group

        sub_group = GenericMetricGroup(
            self,
            name,
            metric_group_type)
        self._sub_groups.append(sub_group)
        return sub_group

    def _get_metric_group_names_and_types(self) -> Tuple[List[str], List[str]]:
        if self._name is None:
            return [], []
        else:
            names, types = self._parent._get_metric_group_names_and_types()
            names.append(self._name)
            types.append(str(self._metric_group_type))
            return names, types

    def _get_namespace(self, time=None) -> str:
        names, metric_group_type = self._get_metric_group_names_and_types()
        names.extend(metric_group_type)
        if time is not None:
            names.append(str(time))
        return json.dumps(names)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.java_gateway import get_gateway

__all__ = ['SqlDialect']


class SqlDialect(object):
    """
    Enumeration of valid SQL compatibility modes.

    In most of the cases, the built-in compatibility mode should be sufficient. For some features,
    i.e. the "INSERT INTO T PARTITION(a='xxx') ..." grammar, you may need to switch to the
    Hive dialect if required.

    We may introduce other SQL dialects in the future.

    :data:`DEFAULT`:

    Flink's default SQL behavior.

    :data:`HIVE`:

    SQL dialect that allows some Apache Hive specific grammar.

    Note: We might never support all of the Hive grammar. See the documentation for
    supported features.
    """

    DEFAULT = 0
    HIVE = 1

    @staticmethod
    def _from_j_sql_dialect(j_sql_dialect):
        gateway = get_gateway()
        JSqlDialect = gateway.jvm.org.apache.flink.table.api.SqlDialect
        if j_sql_dialect == JSqlDialect.DEFAULT:
            return SqlDialect.DEFAULT
        elif j_sql_dialect == JSqlDialect.HIVE:
            return SqlDialect.HIVE
        else:
            raise Exception("Unsupported Java SQL dialect: %s" % j_sql_dialect)

    @staticmethod
    def _to_j_sql_dialect(sql_dialect):
        gateway = get_gateway()
        JSqlDialect = gateway.jvm.org.apache.flink.table.api.SqlDialect
        if sql_dialect == SqlDialect.DEFAULT:
            return JSqlDialect.DEFAULT
        elif sql_dialect == SqlDialect.HIVE:
            return JSqlDialect.HIVE
        else:
            raise TypeError("Unsupported SQL dialect: %s, supported SQL dialects are: "
                            "SqlDialect.DEFAULT, SqlDialect.HIVE." % sql_dialect)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
# cython: language_level = 3
# cython: infer_types = True
# cython: profile=True
# cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
from libc.stdint cimport int32_t, int64_t
from libc.stdlib cimport free, malloc

import datetime
import decimal
import pickle
from typing import List, Union

import cloudpickle
import avro.schema as avro_schema

from pyflink.common import Row, RowKind
from pyflink.common.time import Instant
from pyflink.datastream.window import CountWindow, TimeWindow, GlobalWindow
from pyflink.fn_execution.formats.avro import FlinkAvroDecoder, FlinkAvroDatumReader, \
    FlinkAvroBufferWrapper, FlinkAvroEncoder, FlinkAvroDatumWriter
from pyflink.fn_execution.ResettableIO import ResettableIO
from pyflink.table.utils import pandas_to_arrow, arrow_to_pandas

ROW_KIND_BIT_SIZE = 2

cdef class InternalRow:
    def __cinit__(self, list values, InternalRowKind row_kind):
        self.values = values
        self.row_kind = row_kind
        self.field_names = []

    cpdef object to_row(self):
        row = Row()
        row._values = self.values
        row.set_field_names(self.field_names)
        row.set_row_kind(RowKind(self.row_kind))
        return row

    @staticmethod
    def from_row(row: Row) -> InternalRow:
        cdef InternalRow internal_row
        internal_row = InternalRow(row._values, row.get_row_kind().value)
        internal_row.field_names = row._fields
        return internal_row

    cdef bint is_retract_msg(self):
        return self.row_kind == InternalRowKind.UPDATE_BEFORE or \
           self.row_kind == InternalRowKind.DELETE

    cdef bint is_accumulate_msg(self):
        return self.row_kind == InternalRowKind.UPDATE_AFTER or \
           self.row_kind == InternalRowKind.INSERT

    def __eq__(self, other):
        if not other:
            return False
        return self.values == other.values

    def __getitem__(self, item):
        return self.values[item]

    def __iter__(self):
        return iter(self.values)

    def __repr__(self):
        return "InternalRow(%s, %s)" % (self.row_kind, self.values)

cdef class MaskUtils:
    """
    A util class used to encode mask value.
    """

    def __cinit__(self, field_count):
        self._mask = <bint*> malloc((field_count + ROW_KIND_BIT_SIZE) * sizeof(bint))
        if self._mask == NULL:
            raise MemoryError()

        self._mask_byte_search_table = <unsigned char*> malloc(8 * sizeof(unsigned char))
        if self._mask_byte_search_table == NULL:
            raise MemoryError()
        self._mask_byte_search_table[0] = 0x80
        self._mask_byte_search_table[1] = 0x40
        self._mask_byte_search_table[2] = 0x20
        self._mask_byte_search_table[3] = 0x10
        self._mask_byte_search_table[4] = 0x08
        self._mask_byte_search_table[5] = 0x04
        self._mask_byte_search_table[6] = 0x02
        self._mask_byte_search_table[7] = 0x01

        self._row_kind_byte_table = <unsigned char*> malloc(8 * sizeof(unsigned char))
        if self._row_kind_byte_table == NULL:
            raise MemoryError()
        self._row_kind_byte_table[0] = 0x00
        self._row_kind_byte_table[1] = 0x80
        self._row_kind_byte_table[2] = 0x40
        self._row_kind_byte_table[3] = 0xC0

    def __init__(self, field_count):
        self._field_count = field_count
        self._leading_complete_bytes_num = (self._field_count + ROW_KIND_BIT_SIZE) // 8
        self._remaining_bits_num = (self._field_count + ROW_KIND_BIT_SIZE) % 8

    cdef void write_mask(self, list value, unsigned char row_kind_value,
                         OutputStream output_stream):
        cdef size_t field_pos, index, i
        cdef unsigned char*bit_map_byte_search_table
        cdef unsigned char b
        field_pos = 0
        bit_map_byte_search_table = self._mask_byte_search_table

        # first byte contains the row kind bits
        b = self._row_kind_byte_table[row_kind_value]
        for i in range(0, 8 - ROW_KIND_BIT_SIZE):
            if field_pos + i < self._field_count and value[field_pos + i] is None:
                b |= bit_map_byte_search_table[i + ROW_KIND_BIT_SIZE]
        field_pos += 8 - ROW_KIND_BIT_SIZE
        output_stream.write_byte(b)

        for _ in range(1, self._leading_complete_bytes_num):
            b = 0x00
            for i in range(8):
                if value[field_pos + i] is None:
                    b |= bit_map_byte_search_table[i]
            field_pos += 8
            output_stream.write_byte(b)

        if self._leading_complete_bytes_num >= 1 and self._remaining_bits_num:
            b = 0x00
            for i in range(self._remaining_bits_num):
                if value[field_pos + i] is None:
                    b |= bit_map_byte_search_table[i]
            output_stream.write_byte(b)

    cdef bint*read_mask(self, InputStream input_stream):
        cdef size_t field_pos, i
        cdef unsigned char b
        field_pos = 0
        for _ in range(self._leading_complete_bytes_num):
            b = input_stream.read_byte()
            for i in range(8):
                self._mask[field_pos] = (b & self._mask_byte_search_table[i]) > 0
                field_pos += 1

        if self._remaining_bits_num:
            b = input_stream.read_byte()
            for i in range(self._remaining_bits_num):
                self._mask[field_pos] = (b & self._mask_byte_search_table[i]) > 0
                field_pos += 1
        return self._mask

    def __dealloc__(self):
        if self._mask != NULL:
            free(self._mask)
        if self._mask_byte_search_table != NULL:
            free(self._mask_byte_search_table)
        if self._row_kind_byte_table != NULL:
            free(self._row_kind_byte_table)

cdef class LengthPrefixBaseCoderImpl:
    """
    LengthPrefixBaseCoderImpl will be used in Operations and other coders will be the field coder of
    LengthPrefixBaseCoderImpl.
    """

    def __init__(self, field_coder: FieldCoderImpl):
        self._field_coder = field_coder
        self._data_out_stream = OutputStream()

    cpdef encode_to_stream(self, value, LengthPrefixOutputStream output_stream):
        pass

    cpdef decode_from_stream(self, LengthPrefixInputStream input_stream):
        pass

    cdef void _write_data_to_output_stream(self, LengthPrefixOutputStream output_stream):
        cdef OutputStream data_out_stream
        data_out_stream = self._data_out_stream
        output_stream.write(data_out_stream.buffer, data_out_stream.pos)
        data_out_stream.pos = 0

cdef class FieldCoderImpl:
    cpdef encode_to_stream(self, value, OutputStream out_stream):
        """
        Encodes `value` to the output stream.

        :param value: The output data
        :param out_stream: Output Stream
        """
        pass

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        """
        Decodes data from the input stream.

        :param in_stream: Input Stream
        :param size: The size data of input stream will be decoded. If the size is set 0, it means
        the coder won't take use of the length to decode the data from input stream.
        :return: The decoded object.
        """
        pass

    cpdef bytes encode(self, value):
        """
        Encodes `value` to a bytes

        :param value: The output data
        :return: The encoded data.
        """
        cdef OutputStream out_stream
        out_stream = OutputStream()
        self.encode_to_stream(value, out_stream)
        return out_stream.buffer[0:out_stream.pos]

    cpdef decode(self, encoded):
        """
        Decodes an object from the bytes

        :param encoded: The bytes
        :return: The decoded object.
        """
        cdef InputStream input_stream
        input_stream = InputStream()
        input_stream._input_data = <char*> encoded
        return self.decode_from_stream(input_stream, len(encoded))

cdef class InputStreamWrapper:
    def __cinit__(self, value_coder: ValueCoderImpl, input_stream: LengthPrefixInputStream):
        self._value_coder = value_coder
        self._input_stream = input_stream

    cpdef bint has_next(self):
        return self._input_stream.available()

    cpdef next(self):
        return self._value_coder.decode_from_stream(self._input_stream)

cdef class IterableCoderImpl(LengthPrefixBaseCoderImpl):
    """
    Encodes iterable data to output stream. The output mode will decide whether write a special end
    message 0x00 to output stream after encoding data.
    """

    def __cinit__(self, *args, **kwargs):
        self._end_message = <char*> malloc(1)
        if self._end_message == NULL:
            raise MemoryError()
        self._end_message[0] = 0x00

    def __init__(self, field_coder: FieldCoderImpl, separated_with_end_message: bool):
        super(IterableCoderImpl, self).__init__(field_coder)
        self._separated_with_end_message = separated_with_end_message

    cpdef encode_to_stream(self, value, LengthPrefixOutputStream output_stream):
        if value:
            for item in value:
                self._field_coder.encode_to_stream(item, self._data_out_stream)
                self._write_data_to_output_stream(output_stream)

        # write end message
        if self._separated_with_end_message:
            output_stream.write(self._end_message, 1)

    cpdef decode_from_stream(self, LengthPrefixInputStream input_stream):
        return InputStreamWrapper(ValueCoderImpl(self._field_coder), input_stream)

    def __dealloc__(self):
        if self._end_message != NULL:
            free(self._end_message)

cdef class ValueCoderImpl(LengthPrefixBaseCoderImpl):
    """
    Encodes a single data to output stream.
    """

    def __init__(self, field_coder: FieldCoderImpl):
        super(ValueCoderImpl, self).__init__(field_coder)

    cpdef encode_to_stream(self, value, LengthPrefixOutputStream output_stream):
        self._field_coder.encode_to_stream(value, self._data_out_stream)
        self._write_data_to_output_stream(output_stream)

    cpdef decode_from_stream(self, LengthPrefixInputStream input_stream):
        cdef char*input_data
        cdef char*temp
        cdef size_t size
        cdef long long addr
        cdef InputStream data_input_stream
        # set input_data pointer to the input data
        size = input_stream.read(&input_data)

        # create InputStream
        data_input_stream = InputStream()
        data_input_stream._input_data = input_data

        return self._field_coder.decode_from_stream(data_input_stream, size)

cdef class FlattenRowCoderImpl(FieldCoderImpl):
    """
    A coder for flatten row (List) object (without field names and row kind value is 0).
    """

    def __init__(self, field_coders: List[FieldCoderImpl]):
        self._field_coders = field_coders  # type: List[FieldCoderImpl]
        self._field_count = len(self._field_coders)
        self._mask_utils = MaskUtils(self._field_count)
        self._reuse_flatten_row = [None for i in range(self._field_count)]

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef list list_value, field_coders
        cdef size_t i
        cdef FieldCoderImpl field_coder

        list_value = <list?> value

        # encode mask value
        self._mask_utils.write_mask(list_value, 0, out_stream)

        # encode every field value
        field_coders = self._field_coders
        for i in range(self._field_count):
            item = list_value[i]
            if item is not None:
                field_coder = <FieldCoderImpl> field_coders[i]
                field_coder.encode_to_stream(item, out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef bint*mask
        cdef size_t i
        cdef list flatten_row, field_coders
        cdef FieldCoderImpl field_coder

        # read mask
        mask = self._mask_utils.read_mask(in_stream)
        # skip ROW_KIND_BIT_SIZE to data mask
        (<bint**> &mask)[0] += ROW_KIND_BIT_SIZE

        # decode field data
        flatten_row = self._reuse_flatten_row
        field_coders = self._field_coders
        for i in range(self._field_count):
            if mask[i]:
                flatten_row[i] = None
            else:
                field_coder = <FieldCoderImpl> field_coders[i]
                flatten_row[i] = field_coder.decode_from_stream(in_stream, 0)
        return flatten_row

    def __repr__(self):
        return 'FlattenRowCoderImpl[%s]' % ', '.join(str(c) for c in self._field_coders)

cdef class RowCoderImpl(FieldCoderImpl):
    """
    A coder for `Row` or `InternalRow` object.
    """

    def __init__(self, field_coders: List[FieldCoderImpl], field_names: List[str]):
        self._field_coders = field_coders  # type: List[FieldCoderImpl]
        self._field_count = len(self._field_coders)
        self._field_names = field_names  # type: List[str]
        self._mask_utils = MaskUtils(self._field_count)

    cpdef encode_to_stream(self, value: Union[Row, InternalRow], OutputStream out_stream):
        cdef list list_values
        cdef size_t i
        cdef unsigned char row_kind_value
        cdef FieldCoderImpl field_coder

        if isinstance(value, InternalRow):
            # the type is InternalRow
            list_values = <list> value.values
            row_kind_value = <unsigned char> value.row_kind
        else:
            # the type is Row
            list_values = <list> value.get_fields_by_names(self._field_names)
            row_kind_value = <unsigned char> value.get_row_kind().value
        # encode mask value
        self._mask_utils.write_mask(list_values, row_kind_value, out_stream)

        # encode every field value
        for i in range(self._field_count):
            item = list_values[i]
            if item is not None:
                field_coder = <FieldCoderImpl> self._field_coders[i]
                field_coder.encode_to_stream(item, out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef bint*mask
        cdef unsigned char row_kind_value = 0
        cdef size_t i
        cdef FieldCoderImpl field_coder
        cdef list field_values
        mask = self._mask_utils.read_mask(in_stream)
        field_values = []
        for i in range(self._field_count):
            if mask[i + ROW_KIND_BIT_SIZE]:
                field_values.append(None)
            else:
                field_coder = <FieldCoderImpl> self._field_coders[i]
                field_values.append(field_coder.decode_from_stream(in_stream, 0))
        row = Row()
        row._values = field_values
        for i in range(ROW_KIND_BIT_SIZE):
            row_kind_value += mask[i] * 2 ** i
        row.set_field_names(self._field_names)
        row.set_row_kind(RowKind(row_kind_value))
        return row

    def __repr__(self):
        return 'RowCoderImpl[%s, %s]' % \
               (', '.join(str(c) for c in self._field_coders), self._field_names)

cdef class ArrowCoderImpl(FieldCoderImpl):
    """
    A coder for arrow format data.
    """

    def __init__(self, schema, row_type, timezone):
        self._schema = schema
        self._field_types = row_type.field_types()
        self._timezone = timezone
        self._resettable_io = ResettableIO()
        self._batch_reader = self._load_from_stream(self._resettable_io)

    cpdef encode_to_stream(self, cols, OutputStream out_stream):
        import pyarrow as pa

        self._resettable_io.set_output_stream(out_stream)
        batch_writer = pa.RecordBatchStreamWriter(self._resettable_io, self._schema)
        batch_writer.write_batch(
            pandas_to_arrow(self._schema, self._timezone, self._field_types, cols))

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return self.decode_one_batch_from_stream(in_stream, size)

    cdef list decode_one_batch_from_stream(self, InputStream in_stream, size_t size):
        self._resettable_io.set_input_bytes(in_stream.read(size))
        # there is only one arrow batch in the underlying input stream
        return arrow_to_pandas(self._timezone, self._field_types, [next(self._batch_reader)])

    def _load_from_stream(self, stream):
        import pyarrow as pa

        while stream.readable():
            reader = pa.ipc.open_stream(stream)
            yield reader.read_next_batch()

    def __repr__(self):
        return 'ArrowCoderImpl[%s]' % self._schema

cdef class OverWindowArrowCoderImpl(FieldCoderImpl):
    """
    A coder for over window with arrow format data.
    The data structure: [window data][arrow format data].
    """
    def __init__(self, arrow_coder_impl: ArrowCoderImpl):
        self._arrow_coder = arrow_coder_impl
        self._int_coder = IntCoderImpl()

    cpdef encode_to_stream(self, cols, OutputStream out_stream):
        self._arrow_coder.encode_to_stream(cols, out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef int32_t window_num, window_size
        cdef list window_boundaries_and_arrow_data
        window_num = self._int_coder.decode_from_stream(in_stream, 0)
        size -= 4
        window_boundaries_and_arrow_data = []
        for _ in range(window_num):
            window_size = self._int_coder.decode_from_stream(in_stream, 0)
            size -= 4
            window_boundaries_and_arrow_data.append(
                [self._int_coder.decode_from_stream(in_stream, 0)
                 for _ in range(window_size)])
            size -= 4 * window_size
        window_boundaries_and_arrow_data.append(
            self._arrow_coder.decode_one_batch_from_stream(in_stream, size))
        return window_boundaries_and_arrow_data

    def __repr__(self):
        return 'OverWindowArrowCoderImpl[%s]' % self._arrow_coder


cdef class TinyIntCoderImpl(FieldCoderImpl):
    """
    A coder for tiny int value (from -128 to 127).
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_int8(value)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_int8()

cdef class SmallIntCoderImpl(FieldCoderImpl):
    """
    A coder for small int value (from -32,768 to 32,767).
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_int16(value)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_int16()

cdef class IntCoderImpl(FieldCoderImpl):
    """
    A coder for int value (from -2,147,483,648 to 2,147,483,647).
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_int32(value)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_int32()

cdef class BigIntCoderImpl(FieldCoderImpl):
    """
    A coder for big int value (from -9,223,372,036,854,775,808 to 9,223,372,036,854,775,807).
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_int64(value)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_int64()

cdef class BooleanCoderImpl(FieldCoderImpl):
    """
    A coder for a boolean value.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_byte(value)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return not not in_stream.read_byte()

cdef class FloatCoderImpl(FieldCoderImpl):
    """
    A coder for a float value (4-byte single precision floating point number).
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_float(value)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_float()

cdef class DoubleCoderImpl(FieldCoderImpl):
    """
    A coder for a double value (8-byte double precision floating point number).
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_double(value)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_double()

cdef class BinaryCoderImpl(FieldCoderImpl):
    """
    A coder for a bytes value.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_bytes(value, len(value))

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_bytes()

cdef class CharCoderImpl(FieldCoderImpl):
    """
    A coder for a str value.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef bytes value_bytes = value.encode('utf-8')
        out_stream.write_bytes(value_bytes, len(value_bytes))

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return in_stream.read_bytes().decode("utf-8")

cdef class BigDecimalCoderImpl(FieldCoderImpl):
    """
    A coder for a big decimal value (without fixed precision and scale).
    """

    def __init__(self):
        self._value_coder = CharCoderImpl()

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        self._value_coder.encode_to_stream(str(value), out_stream)

    cpdef decode_from_stream(self, InputStream input_stream, size_t size):
        return decimal.Decimal(self._value_coder.decode_from_stream(input_stream, 0))

cdef class DecimalCoderImpl(FieldCoderImpl):
    """
    A coder for a decimal value (with fixed precision and scale).
    """

    def __init__(self, precision, scale):
        self._context = decimal.Context(prec=precision)
        self._scale_format = decimal.Decimal(10) ** -scale
        self._value_coder = CharCoderImpl()

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        user_context = decimal.getcontext()
        decimal.setcontext(self._context)
        self._value_coder.encode_to_stream(str(value.quantize(self._scale_format)), out_stream)
        decimal.setcontext(user_context)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        user_context = decimal.getcontext()
        decimal.setcontext(self._context)
        result = decimal.Decimal(self._value_coder.decode_from_stream(in_stream, 0)) \
            .quantize(self._scale_format)
        decimal.setcontext(user_context)
        return result

cdef class DateCoderImpl(FieldCoderImpl):
    """
    A coder for a datetime.date value.
    """

    def __init__(self):
        self._EPOCH_ORDINAL = datetime.datetime(1970, 1, 1).toordinal()

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_int32(value.toordinal() - self._EPOCH_ORDINAL)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return datetime.date.fromordinal(in_stream.read_int32() + self._EPOCH_ORDINAL)

cdef class TimeCoderImpl(FieldCoderImpl):
    """
    A coder for a datetime.time value.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef int32_t hour, minute, seconds, microsecond, milliseconds
        hour = value.hour
        minute = value.minute
        seconds = value.second
        microsecond = value.microsecond
        milliseconds = hour * 3600000 + minute * 60000 + seconds * 1000 + microsecond // 1000
        out_stream.write_int32(milliseconds)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef int32_t value, seconds, milliseconds, minutes, hours
        value = in_stream.read_int32()
        seconds = value // 1000
        milliseconds = value % 1000
        minutes = seconds // 60
        seconds %= 60
        hours = minutes // 60
        minutes %= 60
        return datetime.time(hours, minutes, seconds, milliseconds * 1000)

cdef class TimestampCoderImpl(FieldCoderImpl):
    """
    A coder for a datetime.datetime value.
    """

    def __init__(self, precision):
        self._is_compact = precision <= 3

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef int32_t microseconds_of_second, nanoseconds
        cdef int64_t timestamp_seconds, timestamp_milliseconds
        timestamp_seconds = <int64_t> (value.replace(tzinfo=datetime.timezone.utc).timestamp())
        microseconds_of_second = value.microsecond
        timestamp_milliseconds = timestamp_seconds * 1000 + microseconds_of_second // 1000
        nanoseconds = microseconds_of_second % 1000 * 1000
        if self._is_compact:
            out_stream.write_int64(timestamp_milliseconds)
        else:
            out_stream.write_int64(timestamp_milliseconds)
            out_stream.write_int32(nanoseconds)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return self._decode_timestamp_data_from_stream(in_stream)

    cdef _decode_timestamp_data_from_stream(self, InputStream in_stream):
        cdef int64_t milliseconds
        cdef int32_t nanoseconds, seconds, microseconds
        if self._is_compact:
            milliseconds = in_stream.read_int64()
            nanoseconds = 0
        else:
            milliseconds = in_stream.read_int64()
            nanoseconds = in_stream.read_int32()
        seconds = milliseconds // 1000
        microseconds = milliseconds % 1000 * 1000 + nanoseconds // 1000
        return datetime.datetime.utcfromtimestamp(seconds).replace(microsecond=microseconds)

cdef class LocalZonedTimestampCoderImpl(TimestampCoderImpl):
    """
    A coder for a datetime.datetime with time zone value.
    """

    def __init__(self, precision, timezone):
        super(LocalZonedTimestampCoderImpl, self).__init__(precision)
        self._timezone = timezone

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return self._timezone.localize(self._decode_timestamp_data_from_stream(in_stream))

cdef class InstantCoderImpl(FieldCoderImpl):
    """
    A coder for Instant.
    """

    def __init__(self):
        self._null_seconds = -9223372036854775808
        self._null_nanos = -2147483648

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        if value is None:
            out_stream.write_int64(self._null_seconds)
            out_stream.write_int32(self._null_nanos)
        else:
            out_stream.write_int64(value.seconds)
            out_stream.write_int32(value.nanos)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef int64_t seconds
        cdef int32_t nanos
        seconds = in_stream.read_int64()
        nanos = in_stream.read_int32()
        if seconds == self._null_seconds and nanos == self._null_nanos:
            return None
        else:
            return Instant(seconds, nanos)


cdef class CloudPickleCoderImpl(FieldCoderImpl):
    """
    A coder used with cloudpickle for all kinds of python object.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef bytes pickled_bytes
        pickled_bytes = cloudpickle.dumps(value)
        out_stream.write_bytes(pickled_bytes, len(pickled_bytes))

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef bytes pickled_bytes
        pickled_bytes = in_stream.read_bytes()
        return cloudpickle.loads(pickled_bytes)


cdef class PickleCoderImpl(FieldCoderImpl):
    """
    A coder used with pickle for all kinds of python object.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef bytes pickled_bytes
        pickled_bytes = pickle.dumps(value)
        out_stream.write_bytes(pickled_bytes, len(pickled_bytes))

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef bytes pickled_bytes
        pickled_bytes = in_stream.read_bytes()
        return pickle.loads(pickled_bytes)

cdef class GenericArrayCoderImpl(FieldCoderImpl):
    """
    A coder for basic array value (the element of array could be null).
    """

    def __init__(self, elem_coder: FieldCoderImpl):
        self._elem_coder = elem_coder

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef int32_t length, i
        cdef list list_value = value
        length = len(value)
        out_stream.write_int32(length)
        for i in range(length):
            item = list_value[i]
            if item is None:
                out_stream.write_byte(False)
            else:
                out_stream.write_byte(True)
                self._elem_coder.encode_to_stream(item, out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef int32_t length
        length = in_stream.read_int32()
        return [self._elem_coder.decode_from_stream(in_stream, 0) if in_stream.read_byte()
                else None for _ in range(length)]

cdef class PrimitiveArrayCoderImpl(FieldCoderImpl):
    """
    A coder for primitive array value (the element of array won't be null).
    """

    def __init__(self, elem_coder: FieldCoderImpl):
        self._elem_coder = elem_coder

    cpdef encode_to_stream(self, value, OutputStream output_stream):
        cdef int32_t length, i
        cdef list list_value = value
        length = len(value)
        output_stream.write_int32(length)
        for i in range(length):
            self._elem_coder.encode_to_stream(list_value[i], output_stream)

    cpdef decode_from_stream(self, InputStream input_stream, size_t size):
        cdef int32_t length
        length = input_stream.read_int32()
        return [self._elem_coder.decode_from_stream(input_stream, 0) for _ in range(length)]

cdef class MapCoderImpl(FieldCoderImpl):
    """
    A coder for map value (dict with same type key and same type value).
    """

    def __init__(self, key_coder: FieldCoderImpl, value_coder: FieldCoderImpl):
        self._key_coder = key_coder
        self._value_coder = value_coder

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef int32_t length
        length = len(value)
        out_stream.write_int32(length)
        items = value.items()
        for k, v in items:
            self._key_coder.encode_to_stream(k, out_stream)
            if v is None:
                out_stream.write_byte(True)
            else:
                out_stream.write_byte(False)
                self._value_coder.encode_to_stream(v, out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef int32_t length
        cdef dict map_value
        length = in_stream.read_int32()
        map_value = {}
        for _ in range(length):
            key = self._key_coder.decode_from_stream(in_stream, 0)
            if in_stream.read_byte():
                map_value[key] = None
            else:
                map_value[key] = self._value_coder.decode_from_stream(in_stream, 0)
        return map_value

cdef class TupleCoderImpl(FieldCoderImpl):
    """
    A coder for a tuple value.
    """

    def __init__(self, field_coders):
        self._field_coders = field_coders
        self._field_count = len(field_coders)

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        cdef tuple tuple_value = value
        for i in range(self._field_count):
            self._field_coders[i].encode_to_stream(tuple_value[i], out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef int32_t i
        cdef list decoded_list
        decoded_list = [self._field_coders[i].decode_from_stream(in_stream, 0)
                        for i in range(self._field_count)]
        return tuple(decoded_list)

cdef class TimeWindowCoderImpl(FieldCoderImpl):
    """
    A coder for TimeWindow.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_int64(value.start)
        out_stream.write_int64(value.end)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        cdef int64_t start, end
        start = in_stream.read_int64()
        end = in_stream.read_int64()
        return TimeWindow(start, end)

cdef class CountWindowCoderImpl(FieldCoderImpl):
    """
    A coder for CountWindow.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_int64(value.id)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return CountWindow(in_stream.read_int64())

cdef class GlobalWindowCoderImpl(FieldCoderImpl):
    """
    A coder for GlobalWindow.
    """

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        out_stream.write_byte(0)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        in_stream.read_byte()
        return GlobalWindow()

cdef class DataViewFilterCoderImpl(FieldCoderImpl):
    """
    A coder for CountWindow.
    """
    def __init__(self, udf_data_view_specs):
        self._udf_data_view_specs = udf_data_view_specs
        self._pickle_coder = PickleCoderImpl()

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        self._pickle_coder.encode_to_stream(self._filter_data_views(value), out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        return self._pickle_coder.decode_from_stream(in_stream, size)

    def _filter_data_views(self, row):
        i = 0
        for specs in self._udf_data_view_specs:
            for spec in specs:
                row[i][spec.field_index] = None
            i += 1
        return row

cdef class AvroCoderImpl(FieldCoderImpl):

    def __init__(self, schema_string: str):
        self._buffer_wrapper = FlinkAvroBufferWrapper()
        self._schema = avro_schema.parse(schema_string)
        self._decoder = FlinkAvroDecoder(self._buffer_wrapper)
        self._encoder = FlinkAvroEncoder(self._buffer_wrapper)
        self._reader = FlinkAvroDatumReader(writer_schema=self._schema, reader_schema=self._schema)
        self._writer = FlinkAvroDatumWriter(writer_schema=self._schema)

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        self._buffer_wrapper.switch_stream(out_stream)
        self._writer.write(value, self._encoder)

    cpdef decode_from_stream(self, InputStream in_stream, size_t size):
        self._buffer_wrapper.switch_stream(in_stream)
        return self._reader.read(self._decoder)

cdef class LocalDateCoderImpl(FieldCoderImpl):

    @staticmethod
    def _encode_to_stream(value, OutputStream out_stream):
        if value is None:
            out_stream.write_int32(0xFFFFFFFF)
            out_stream.write_int16(0xFFFF)
        else:
            out_stream.write_int32(value.year)
            out_stream.write_int8(value.month)
            out_stream.write_int8(value.day)

    @staticmethod
    def _decode_from_stream(InputStream in_stream):
        year = in_stream.read_int32()
        if year == 0xFFFFFFFF:
            in_stream.read(2)
            return None
        month = in_stream.read_int8()
        day = in_stream.read_int8()
        return datetime.date(year, month, day)

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        return self._encode_to_stream(value, out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t length):
        return self._decode_from_stream(in_stream)

cdef class LocalTimeCoderImpl(FieldCoderImpl):

    @staticmethod
    def _encode_to_stream(value, OutputStream out_stream):
        if value is None:
            out_stream.write_int8(0xFF)
            out_stream.write_int16(0xFFFF)
            out_stream.write_int32(0xFFFFFFFF)
        else:
            out_stream.write_int8(value.hour)
            out_stream.write_int8(value.minute)
            out_stream.write_int8(value.second)
            out_stream.write_int32(value.microsecond * 1000)

    @staticmethod
    def _decode_from_stream(InputStream in_stream):
        hour = in_stream.read_int8()
        if hour == 0xFF:
            in_stream.read(6)
            return None
        minute = in_stream.read_int8()
        second = in_stream.read_int8()
        nano = in_stream.read_int32()
        return datetime.time(hour, minute, second, nano // 1000)

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        return self._encode_to_stream(value, out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t length):
        return self._decode_from_stream(in_stream)

cdef class LocalDateTimeCoderImpl(FieldCoderImpl):

    cpdef encode_to_stream(self, value, OutputStream out_stream):
        if value is None:
            LocalDateCoderImpl._encode_to_stream(None, out_stream)
            LocalTimeCoderImpl._encode_to_stream(None, out_stream)
        else:
            LocalDateCoderImpl._encode_to_stream(value.date(), out_stream)
            LocalTimeCoderImpl._encode_to_stream(value.time(), out_stream)

    cpdef decode_from_stream(self, InputStream in_stream, size_t length):
        date = LocalDateCoderImpl._decode_from_stream(in_stream)
        time = LocalTimeCoderImpl._decode_from_stream(in_stream)
        if date is None or time is None:
            return None
        return datetime.datetime(date.year, date.month, date.day, time.hour, time.minute,
                                 time.second, time.microsecond)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from apache_beam.coders import Coder
from apache_beam.coders.coders import FastCoder, LengthPrefixCoder
from apache_beam.portability import common_urns
from apache_beam.typehints import typehints

from pyflink import fn_execution
from pyflink.fn_execution.coders import LengthPrefixBaseCoder
from pyflink.fn_execution.flink_fn_execution_pb2 import CoderInfoDescriptor

if fn_execution.PYFLINK_CYTHON_ENABLED:
    from pyflink.fn_execution.beam import beam_coder_impl_fast as beam_coder_impl
    from pyflink.fn_execution.beam.beam_coder_impl_fast import FlinkFieldCoderBeamWrapper
    from pyflink.fn_execution.beam.beam_coder_impl_fast import FlinkLengthPrefixCoderBeamWrapper
else:
    from pyflink.fn_execution.beam import beam_coder_impl_slow as beam_coder_impl
    from pyflink.fn_execution.beam.beam_coder_impl_slow import FlinkFieldCoderBeamWrapper
    from pyflink.fn_execution.beam.beam_coder_impl_slow import FlinkLengthPrefixCoderBeamWrapper

FLINK_CODER_URN = "flink:coder:v1"


class PassThroughLengthPrefixCoder(LengthPrefixCoder):
    """
    Coder which doesn't prefix the length of the encoded object as the length prefix will be handled
    by the wrapped value coder.
    """

    def __init__(self, value_coder):
        super(PassThroughLengthPrefixCoder, self).__init__(value_coder)

    def _create_impl(self):
        return beam_coder_impl.PassThroughLengthPrefixCoderImpl(self._value_coder.get_impl())

    def __repr__(self):
        return 'PassThroughLengthPrefixCoder[%s]' % self._value_coder


Coder.register_structured_urn(
    common_urns.coders.LENGTH_PREFIX.urn, PassThroughLengthPrefixCoder)


class FlinkCoder(FastCoder):

    def __init__(self, internal_coder):
        self._internal_coder = internal_coder

    def _create_impl(self):
        return self._internal_coder.get_impl()

    def get_impl(self):
        if isinstance(self._internal_coder, LengthPrefixBaseCoder):
            return FlinkLengthPrefixCoderBeamWrapper(self._create_impl())
        else:
            return FlinkFieldCoderBeamWrapper(self._create_impl())

    def to_type_hint(self):
        return typehints.Any

    @Coder.register_urn(FLINK_CODER_URN, CoderInfoDescriptor)
    def _pickle_from_runner_api_parameter(
            coder_info_descriptor_proto, unused_components, unused_context):
        return FlinkCoder(LengthPrefixBaseCoder.from_coder_info_descriptor_proto(
            coder_info_descriptor_proto))

    def __repr__(self):
        return 'FlinkCoder[%s]' % repr(self._internal_coder)

    def __eq__(self, other: 'FlinkCoder'):
        return (self.__class__ == other.__class__
                and self._internal_coder == other._internal_coder)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self._internal_coder)
